/*
** Module   : WebClientBuilderOptions.java
** Abstract : store chuiWeb parameters read from directory.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140206 First version, extracted from ClientBuilderOptions.
** 002 MAG 20140711 Enable/Disable clipboard.
** 003 GES 20150310 Made this class generic for web clients, not just for ChUI web clients.
*/

package com.goldencode.p2j.main;

import java.util.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;

/**
 * The parameters stored inside this structure are read from directory on server start-up 
 * and are used as default parameters to create and spawn a web client process.
 * <p>
 * This class MUST NOT BE USED from client side, as it uses a direct access to the server.  Also,
 * this class can be instantiated only from threads with a P2J context for the running P2J server.
 */
public class WebClientBuilderOptions
extends ClientBuilderOptions
{   
   /** Directory node name in which our configuration can be found. */
   public static final String DIRECTORY_NODE_ID = "webClient";

   /**
    * Singleton instance. Not set as final to allow it to be invalidated (in case these
    * options are changed via admin console).
    */
   private static WebClientBuilderOptions instance = null;

   /** Options storage. */
   private final Map<String, String> options = new HashMap<>();

   /**
    * Initialize. Read all configuration parameters from directory.
    * 
    * @throws   IllegalStateException
    *           If this is called from P2J clients.
    */
   private WebClientBuilderOptions()
   {
      super(null);
      
      if (SessionManager.get() == null || SessionManager.get().isLeaf())
      {
         throw new IllegalStateException("This class can be used only on the server-side!");
      }

      // read the configuration from the directory
      readOptions();
   }
   
   /**
    * Get the {@link WebClientBuilderOptions} singleton instance or initialize it on first
    * access.
    *
    * @return   The singleton instance.
    */
   public static synchronized WebClientBuilderOptions getInstance()
   {
      // TODO: this will not allow changing the options via the AdminConsole - add a way to 
      // invalidate and recreate this instance

      // this needs to be initialized always in a thread with the server's context

      if (instance == null)
      {
         instance = new WebClientBuilderOptions();
      }
      
      return instance;
   }

   /**
    * Get web client driver options. This contains configuration for both GUI and ChUI web
    * clients.  It is up to the caller to use only those portions that make sense for the
    * client type in use.
    * 
    * @return   The web client options.
    */
   public Map<String, String> getClientOptions()
   {
      return options;
   }

   /**
    * Get the authentication mode for this spawner.
    * 
    * @return   Always {@code true}.
    */
   protected boolean isPasswordAuthentication()
   {
      return true;
   }

   /**
    * Read all web client driver options from the directory and store the defaults and any
    * overridden configuration in a map for configuration of both GUI and ChUI web drivers.
    * It is up to the downstream user of the map to use only those portions that make sense
    * for the client type in use.
    */
   private void readOptions()
   {
      // default terminal options
      int     rows       = TerminalOptions.DEF_ROWS;
      int     columns    = TerminalOptions.DEF_COLUMNS;
      
      // for colors, don't include any alpha on bits 24-31, if emitted as FF the result is not
      // interpreted properly by the browser
      String  background = TerminalOptions.DEF_BACKGROUND.asHexString();
      String  foreground = TerminalOptions.DEF_FOREGROUND.asHexString();
      String  selection  = TerminalOptions.DEF_SELECTION.asHexString();
      
      String  fontname   = TerminalOptions.DEF_FONTNAME;
      int     fontsize   = TerminalOptions.DEF_FONTSIZE;
      long    stimeout   = -1;
      long    wtimeout   = -1;
      int     port       = 0;
      String  host       = null;
      boolean clipboard  = true; 

      // get registered terminal options
      rows       = getNode(DIRECTORY_NODE_ID, "rows", rows);
      columns    = getNode(DIRECTORY_NODE_ID, "columns", columns);
      background = getNode(DIRECTORY_NODE_ID, "background", background);
      foreground = getNode(DIRECTORY_NODE_ID, "foreground", foreground);
      selection  = getNode(DIRECTORY_NODE_ID, "selection", selection);
      fontname   = quotedString(getNode(DIRECTORY_NODE_ID, "fontname", fontname));
      fontsize   = getNode(DIRECTORY_NODE_ID, "fontsize", fontsize);
      stimeout   = getNode(DIRECTORY_NODE_ID, "webSocketTimeout", (int) stimeout);
      wtimeout   = getNode(DIRECTORY_NODE_ID, "watchdogTimeout",  (int) wtimeout);
      port       = getNode(DIRECTORY_NODE_ID, "port", port);
      host       = getNode(DIRECTORY_NODE_ID, "host", host);
      clipboard  = getNode(DIRECTORY_NODE_ID, "clipboard", true);
   
      // build client options map
      options.put("rows", String.valueOf(rows));
      options.put("columns", String.valueOf(columns));
      options.put("background", background);
      options.put("foreground", foreground);
      options.put("selection", selection);
      options.put("fontname", fontname);
      options.put("fontsize", String.valueOf(fontsize));
      options.put("webSocketTimeout", String.valueOf(stimeout));
      options.put("watchdogTimeout", String.valueOf(stimeout));
      options.put("port", String.valueOf(port));
      options.put("host", host);
      options.put("clipboard", clipboard ? "true" : "false");
   }
}
