/*
** Module   : ColorPalette.java
** Abstract : Declares the supported terminal colors.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20140119 Created initial version.
** 002 MAG 20140211 Map entries to AWT Color.
** 003 MAG 20140821 Add new color entries for Windows OS.  Provide conversion method for color 
*                   number to palette index and from palette index to color number. 
*/

package com.goldencode.p2j.ui;

/**
 * Set of supported terminal colors.
 * <p>
 * The implicit ordinal created by Java for this enum directly maps 
 * each enum "element" to the "COLOR_element" constant in NCurses/Curses. 
 * This mapping is implicitly utilized in NCurses/Curses drivers without 
 * any further translation. Other drivers will have to provide some kind 
 * of mapping from these values to their native rendering. 
 * <p>
 * WARNING: Do not change the order of enum items.
 * <p>
 * The color palette is implemented as a Java enum type.  The first 8 entries in color palette
 * are used on Linux / Unix only.  On Windows all 16 entries are used. The color phrase should
 * works on both ChUI and GUI interfaces.  On Windows GUI the colors are indexes in COLOR-TABLE
 * structure. Because the ordinal numbers of palette enum are different than color numbers
 * conversion methods has been designed to do the conversions from color number to palette index
 * and from palette index to color number.
 * <p>
 * Conversion Table:
 * <p>
 * <pre>
 * -----------------------------------------------------------------------------------------------
 * Index Color          Abbreviation            Palette     Color          dos-hex    Color
 * -----------------------------------------------------------------------------------------------
 *  0    Black          Bla, Blk                 0          BLACK          0          Black
 *  1    Blue           Blu                     11          DARK_BLUE      1          Blue
 *  2    Green          Gre, Grn                 9          DARK_GREEN     2          Green
 *  3    Cyan           C                       13          DARK_CYAN      3          Aqua
 *  4    Red            Red                      8          DARK_RED       4          Red
 *  5    Magenta        Ma                      12          DARK_MAGENTA   5          Purple
 *  6    Brown          Bro, Brn                10          BROWN          6          Yellow
 *  7    Gray           Gra, Gry                14          GRAY           7          White
 *  8    Dark-Gray      D-Gra                   15          DARK_GRAY      8          Gray
 *  9    Light-Blue     Lt-Blu                   4          BLUE           9          Light Blue
 * 10    Light-Green    Lt-Gre                   2          GREEN          A          Light Green
 * 11    Light-Cyan     Lt-C                     6          CYAN           B          Light Green
 * 12    Light-Red      Lt-Red                   1          RED            C          Light Green
 * 13    Light-Magenta  Lt-Ma                    5          MAGENTA        D          Light Purple
 * 14    Yellow         Y Light-Brown Lt-Br      3          YELLOW         E          Light Yellow
 * 15    White          W                        7          WHITE          F          Bright White
 * -----------------------------------------------------------------------------------------------
 * </pre>
 */

/**
 * ChUI color palette implementation.
 */
public enum ColorPalette
{   
   BLACK(0, 0, 0, 0),
   RED(12, 255, 0, 0),
   GREEN(10, 0, 255, 0),
   YELLOW(14, 255, 255, 0),
   BLUE(9, 0, 0, 255),
   MAGENTA(13, 255, 0, 255),
   CYAN(11, 0, 255, 255),
   WHITE(15, 255, 255, 255),
   DARK_RED(4, 128, 0, 0),
   DARK_GREEN(2, 0, 128, 0),
   BROWN(6, 128, 128, 0),
   DARK_BLUE(1, 0, 0, 128),                          
   DARK_MAGENTA(5, 128, 0, 128),
   DRAK_CYAN(3, 0, 128, 128),
   GRAY(7, 128, 128, 128),
   DARK_GRAY(8, 192, 192, 192);
         
   /** Conversion table color number to palette index */
   private static int[] paletteIndexes;

   /** Conversion table palette index to color number */
   private static int[] colorIndexes;

   /** Color number (index in COLOR-TABLE) */
   private int index;
   
   /** Red. */
   private int r;

   /** Green. */
   private int g;
   
   /** Blue. */
   private int b;
   
   static
   {
      ColorPalette[] values = values();
      
      paletteIndexes = new int[values.length];
      colorIndexes = new int[values.length];
      
      for (int i = 0; i < values.length; i++)
      {
         int colorNum = values[i].getIndex();
         paletteIndexes[colorNum] = i;
         colorIndexes[i] = colorNum;
      }
   }
   
   /**
    * Convert color number to palette index.
    * 
    * @param    color
    *           Color number. Must be between 0 and 15.
    *           
    * @return   Palette index.
    */
   public static int colorToPalette(int color)
   {
      return paletteIndexes[color & 0x0F];
   }

   /**
    * Convert color number to palette index.
    * 
    * @param    index
    *           Palette index. Must be between 0 and 15.
    *                                                 
    * @return   Color number.
    */
   public static int paletteToColor(int index)
   {
      return colorIndexes[index & 0x0F];
   }
   
   /**
    * Constructor.
    * 
    * @param    index
    *           Color index in COLOR-TBALE.
    * @param    r
    *           Red component.
    * @param    g
    *           Green component.
    * @param    b
    *           Blue component
    */
   private ColorPalette(int index, int r, int g, int b)
   {
      this.index = index;
      this.r     = r;
      this.g     = g;
      this.b     = b;
   }
                                                             
   /**
    * Get color index number in COLOR-TABLE.
    * 
    * @return   The color-table index.
    */
   public int getIndex()
   {
      return index;
   }

   /**
    * Get the red component of this color.
    * 
    * @return   The red component.
    */
   public int getRed()
   {
      return r;
   }

   /**
    * Get the green component of this color.
    * 
    * @return   The green component.
    */
   public int getGreen()
   {
      return g;
   }

   /**
    * Get the blue component of this color.
    *                                                        
    * @return   The blue component.
    */
   public int getBlue()
   {
      return b;
   }
}
