/*
** Module   : WebPageHandler.java
** Abstract : delivers the main page for a web client
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 GES 20150312 Created initial version by moving code from the ChuiWebPageHandler and making
**                  it more generic (to handle GUI too).
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.apache.commons.codec.digest.*;
import org.eclipse.jetty.server.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.web.*;
import com.google.common.net.*;

/**
 * Jetty Handler which delivers HTML resource files. This class provide a method to parse the
 * HTML text and substitute place holders. 
 */
public class WebPageHandler
extends HtmlResourceHandler
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WebPageHandler.class.getName());
      
   /** Configuration. */
   private BootstrapConfig config;
   
   /** Cookie token */
   private String cookieToken;   
   
   /** Authorization token */
   private String authorizationToken;   
   
   /**
    * Constructor.
    *
    * @param    config
    *           Configuration parameters.
    * @param    root
    *           Prefix for all URLs relating to this page.
    * @param    skel
    *           Fully qualified resource name of the HTML file which should be used.
    */
   public WebPageHandler(BootstrapConfig config, String root, String skel)
   {
      super(root, skel);
      this.config = config;
            
      LOG.logp(Level.INFO,
               "WebPageHandler",
               "",
               LogHelper.generate("target=%s page=%s", root, skel));
   }
   
   /**
    * Return the main page for AJAX clients. If authorization required and not authorized
    * do not load the main page client will get a HTTP 404 error, page not found.
    *    
    * @param   target 
    *          The target of the request - either a URI or a name.
    * @param   base
    *          The base request.
    * @param   request
    *          The object or a wrapper of the request.
    * @param   response 
    *          The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      if (isAuthorized(base, response))
      {
         super.handle(target, base, request, response);
      }    
   }
   
   /**
    * Set authorization token.
    * 
    * @param   token
    *          Authorization token;
    */
   public void setAuthorizationToken(String token)
   {
      this.authorizationToken = token;
   }
   
   /**
    * Convert to HTML color format.
    * 
    * @param   color
    *          Color text string.
    * @return  HTML format color
    */
   public String htmlColor(String color)
   {
      return color.startsWith("0x") ? color.replace("0x", "#") : color;
   }

   /**
    * Process the given string and make any replacements of parameter
    * values as needed. The replacement placeholder follow the JEE 
    * expression language (EL) format for immediate evaluation. 
    * <p>
    * Example: ${expression} 
    *
    * @param    base
    *           The details and state the request.
    * @param    text
    *           The text line that may need modifications.
    *
    * @return   The new text if a modification occurred or the original text
    *           if no modification was needed.
    */
   protected String handleReplacements(Request base, String text)
   {
      if (text.contains("${context}"))
      {
         text = text.replace("${context}", base.getContextPath());
      }
      
      if (text.contains("${referrer}"))
      {
         String referrer = config.getString("web", "referrer", "url", null);               
         text = text.replace("${referrer}", referrer);
      }
      
      if (text.contains("${clipboard.enabled}"))
      {
         boolean enabled = config.getBoolean("client", "web", "clipboard", true);               
         text = text.replace("${clipboard.enabled}", enabled ? "true" : "false");
      }

      return text;
   }
   
   /**
    * Check authorization. First check the query parameter PARAM_TOKEN. If found and equals with
    * stored token set a new authorization cookie if not already set. This should happen on first
    * connection when the user is redirected.
    * <p>  
    * If query parameter is missing check the browser cookies list for a cookie having an 
    * authorization token that match the expected cookie. If found the request is authorized 
    * otherwise the request is denied. This should happen after initial redirection whenever the
    * page is reloaded.
    * 
    * @param   request
    *          HttpServletRequest instance.
    * @param   response
    *          HttpServletResponse instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   private boolean isAuthorized(HttpServletRequest request, HttpServletResponse response)
   {
      if (authorizationToken == null)
      {
         return false;
      }

      return (cookieToken == null) ? checkQueryParam(request, response) : checkCookies(request);
   }
   
   /**
    * Check authorization token PARAM_TOKEN. If found and equals with stored token set a new
    * authorization cookie if not already was set. This should happen on first connection when
    * the user is redirected. After new cookie is set the authorization token is no longer valid.
    * 
    * @param   request
    *          HttpServletRequest instance.
    * @param   response
    *          HttpServletResponse instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   @SuppressWarnings("deprecation")
   private boolean checkQueryParam(HttpServletRequest request, HttpServletResponse response)
   {
      // check query parameter
      String token = request.getParameter(WebHandler.PARAM_TOKEN);
      
      // compare tokens
      boolean match = authorizationToken.equals(token);
      
      if (match)
      {
         // set a cookie using a new generated token
         cookieToken = DigestUtils.shaHex(UUID.randomUUID().toString());
         Cookie cookie = new Cookie(HttpHeaders.AUTHORIZATION, cookieToken);
         response.addCookie(cookie); 
      }
      
      return match;
   }
   
   /**
    * Check cookies list for a cookie that match the expected value. If found the request is
    * authorized otherwise the request is denied.
    * 
    * @param   request
    *          HttpServletRequest instance.
    *          
    * @return  <code>true</code> if authorized <code>false</code> otherwise.
    */
   private boolean checkCookies(HttpServletRequest request)
   {
      // check browser cookies
      Cookie[] cookies = request.getCookies();
   
      if (cookies != null)
      {
         // search for authorization cookie
         for (Cookie cookie : cookies)
         {
            String name = cookie.getName();
            String value = cookie.getValue();
            
            if (HttpHeaders.AUTHORIZATION.equals(name) && cookieToken.equals(value))
            {
               return true;
            }
         }
      }
      
      return false;      
   }
}
