/*
** Module   : p2j.clipboard.js
** Abstract : core clipboard implementation.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140701 First version.
** 002 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients. Client-specific portions have
**                  been moved to p2j.clipboard_helpers which will have different versions for
**                  ChUI and GUI.
*/

/**
 * Core clipboard implementation that calls client-specific callbacks for ChUI or GUI selection
 * and rendering.  The user drags the mouse to select text and the hooks are called to capture
 * selection positions and to render the selection highlight. After the mouse drag is complete
 * any CTRL-C will copy the text to the clipboard. ESC key will clear the selection. For paste,
 * the CTRL-V key combination is monitored and text read from the clipboard is sent to the server
 * as input. 
 */
p2j.clipboard = (function() 
{
   /** Clipboard object which will be returned from this initialization function. */
   var me = {};
   
   /** CTRL-V pressed flag. */
   me.ctrlv = false;
   
   /** Clipboard enabled/disabled flag. */
   me.enabled = false;
   
   /** Have selection flag. */
   var selection = false;
   
   /** Stroke rectangle selection. */
   var stroke = true;
   
   /** Selection color. */
   var color = 'blue';

   /** Drag flag. */
   var drag = false;
      
   /** Hidden input element used on copy/paste operations. */
   var inputText;
   
   /**
    * Init module.
    * 
    * @param {type} cfg configurartion object
    */
   me.init = function(cfg)
   {
      me.enabled = cfg.clipboard.enabled;
      
      // input element used on copy/paste operations.
      inputText = document.getElementById(cfg.clipboard.input);
      
      p2j.clipboard_helpers.addEventListener('mousedown', mouseDown);
      p2j.clipboard_helpers.addEventListener('mouseup', mouseUp);
      p2j.clipboard_helpers.addEventListener('mousemove', mouseMove);

      require(["dijit/Menu", "dijit/MenuItem", "dijit/CheckedMenuItem", "dijit/MenuSeparator", "dijit/PopupMenuItem"], 
      function(Menu, MenuItem, CheckedMenuItem, MenuSeparator, PopupMenuItem)
      {   
         pMenu = new dijit.Menu(
         {
            targetNodeIds:[cfg.clipboard.id]
         });

         // select 
         pMenuSelect = new dijit.CheckedMenuItem(
         {
            iconClass:"dijitEditorIcon dijitEditorIconSelectAll",
            label: "Enable Clipboard",
            disabled: false,
            checked: me.enabled,
            onClick: function()
            {
               me.enabled = this.checked;
               me.clearSelection();
            }
         });
         
         pMenu.addChild(pMenuSelect);
         pMenu.startup();      
      });
      
      // firefox
      if (!window.Clipboard)
      {
         inputText.focus();
      }
   };
   
   /**
    * Mouse down event handler.
    * 
    * @param {type} e Event object
    * @returns {undefined}
    */
   function mouseDown(e) 
   {
      if (e.button === 0)
      {
         p2j.clipboard_helpers.mouseDown(e);
         drag = true;
      }
   }

   /**
    * Mouse up event handler. Draw selection rectangle.
    * 
    * @param {type} e Event object
    * @returns {undefined}
    */
   function mouseUp(e) 
   {
     if (drag && e.button === 0)
     {
        drag = false;
     }
   }

   /**
    * Mouse move event handler.
    * 
    * @param {type} e Event object
    * @returns {undefined}
    */
   function mouseMove(e) 
   {
      if (me.enabled && drag && e.button === 0) 
      {
         p2j.clipboard_helpers.mouseMove(e);
         selection = true;
      }
   }      

   /**
    * Clipboard paste handler.
    * 
    * @param {type} e event object
    * @returns {undefined}
    */
   window.addEventListener("paste", function pasteHandler(e)
   {
      if (!me.enabled)
         return;
      
      var text = "";
 
      if (window.clipboardData) 
      {   
         text = window.clipboardData.getData("Text");
      }
      else if (e.clipboardData) 
      {
         text = e.clipboardData.getData("text/plain");
      }

      if (isEmpty(text))
      {
         // if empty string and CTRL-V was pressed       
         if (me.ctrlv)
         {
            // send CTRL-V
            var msg = new Uint8Array(3);
            msg[0] = 0x03;
            msg[1] = 0;
            msg[2] = 22;
            p2j.socket.send(msg);
         }
      }
      else      
         paste(text);
      
      me.ctrlv = false;
      e.preventDefault();
   });
   
   /**
    * Send paste text down to server to be injected
    * into the keyboard queue. 
    * 
    * @param {type} text   text to send.
    * @returns {undefined}
    */
   function paste(text)
   {
      if (text.length > 0)
      {
         var msg = new Uint8Array(text.length * 2 + 1);
         msg[0] = 0x03;

         for (var i = 0, j = 1;i < text.length;i++, j += 2)
         {
            var c = text.charCodeAt(i);
            msg[j + 1] = c & 0xff;
            c = c >> 8;
            msg[j] = c & 0xff;
         }
         
         p2j.socket.send(msg); 
      }                     
   }

   /**
    * Clipboard copy handler.
    * 
    * @param {type} e event object
    * @returns {undefined}
    */
   window.addEventListener("copy", function copyHandler(e)
   {
      if (!me.enabled)
         return;

      // value to copy
      var text = inputText.value;
      
      // copy to clipboard
      if (window.clipboardData) 
      {
         window.clipboardData.clearData();
         window.clipboardData.setData("Text", text);
      }
      else if (e.clipboardData) 
      {
         e.clipboardData.clearData();
         e.clipboardData.setData("text/plain", text);
      }      
      
      // clear selection
      me.clearSelection();
      e.preventDefault();
   });
   
   /**
    * Prepare for copy operation.
    * 
    * @returns {undefined}
    */
   me.prepareCopy = function()
   {
      if (selection)
      {
         // get selected text
         var text = p2j.clipboard_helpers.getSelectedText();
         
         // copy selection to clipboard
         inputText.value = text;
         inputText.focus();
         inputText.select();
      }
   };
      
   /**
    * Prepare for paste operation.
    * 
    * @returns {undefined}
    */
   me.preparePaste = function()
   {
      inputText.select();
   };

   /**
    * Check if a string is empty.
    * 
    * @param {str} input string 
    * @returns true if empty false otherwise
    */
   function isEmpty(str) 
   {
      return (!str || 0 === str.length);
   }   
   
   /**
    * Clear selection on screen.
    * 
    * @returns {undefined}
    */
   me.clearSelection = function()
   {
      selection = false;
      p2j.clipboard_helpers.clearSelection();
   };
   
   /**
    * Check if have selection.
    * 
    * @returns {boolean} true if have a selection false if not.
    */
   me.haveSelection = function()
   {
      return selection;
   };

   return me;   
}());
