/*
** Module   : CaptionButton.java
** Abstract : Define caption button associated with a window.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140917 Changed default BGCOLOR.
** 006 CA  20141021 Fixed drawing of caption buttons; allow restore button to show if window is
**                  maximized.
** 007 MAG 20141029 Added clipping rectangle when drawing.
** 008 HC  20150510 Improved drawing of caption buttons - they are now more Windows alike. 
** 009 GES 20150507 Removed AWT dependencies. 
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * A caption button associated with a window.
 */
@SuppressWarnings("unchecked")
class CaptionButton
extends Button<GuiOutputManager>
{
   /** Base width. */
   private final static int BASE_WIDTH = 19;
   
   /** Base height. */
   private final static int BASE_HEIGHT = 17;
   
   /** The button type. */
   private CaptionButtonType type;

   /** The graphics driver. */
   private GuiDriver gd;

   /** The button state - up or down. */
   private boolean up;
   
   /** The button's width, in character units. */
   private double width;
   
   /** The button's height, in character units. */
   private double height;
   
   /** Horizontal drawing scale. */
   private double xScale = 1;
   
   /** Vertical drawing scale. */
   private double yScale = 1;

   /**
    * Create a new button with the specified type.
    * 
    * @param    id
    *           The widget's ID.
    * @param    type
    *           The button type.
    */
   CaptionButton(WidgetId id, CaptionButtonType type)
   {
      super(id, null, "");

      gd = (GuiDriver) OutputManager.getDriver();

      this.type = type;
      this.up = true;
   }
   
   /**
    * Returns the button type.
    * 
    * @return   The button type.
    */
   public CaptionButtonType getType()
   {
      return type;
   }

   /**
    * Draw this caption button, depending on its type.
    */
   @Override
   public void draw()
   {
      NativePoint p = physicalLocation();
      final WindowGuiImpl window = (WindowGuiImpl) ancestor();
      
      NativeDimension d = physicalDimension();
      
      updateScale(d);
      
      final int width = d.width;
      final int height = d.height;

      final NativeRectangle clip = clipRectangle(p, d);
      
      // draw button
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            ColorRgb faceColor = up ? gd.getSysColor(SysColor.COLOR_3DFACE) :
                                      gd.getSysColor(SysColor.COLOR_3DSHADOW);
            
            gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
            gd.fill3DRect(0, 0, width, height, up);
            
            gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
            
            // draw icon on button
            switch (type)
            {   
               case CLOSE:
                  gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, scaleX(2));
                  
                  // north-west - south-east line
                  gd.drawLine(scaleX(4), scaleY(3), width - scaleX(4), height - scaleY(3));
                  
                  // SW-NE line
                  gd.drawLine(scaleX(4), height - scaleY(3), width - scaleX(4), scaleY(3));
                  
                  // TODO: the SW-NE line is painted with 1/2 thickness of the NW-SE line,
                  //       this unfortunately limits the scaling capabilities
                  gd.drawLine(scaleX(5), height - scaleY(3), width - scaleX(3), scaleY(3));

                  // these rects make the proper line ends
                  gd.setColor(faceColor);
                  gd.fillRect(2, 2, width - 4, scaleY(2));  
                  gd.fillRect(2, height - 3, width - 4, scaleY(2));
                  gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));

                  gd.setLineStroke(LineStroke.DEFAULT);
                  break;
               
               case ICONIFY:
                  gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, scaleX(2));
                  
                  gd.drawLine(scaleX(5), height - scaleY(4), width - scaleX(7), height - scaleY(4));
                  gd.setLineStroke(LineStroke.DEFAULT);
                  break;
               
               case MAXIMIZE:
                  if (!window.isMaximized())
                  {
                     drawMaxRect(scaleX(4), 
                                 scaleY(3), 
                                 width - scaleX(4) - scaleX(5), 
                                 height - scaleY(3) - scaleY(4),
                                 false,
                                 faceColor);
                  }
                  else
                  {
                     // draw the upper window icon
                     
                     drawMaxRect(scaleX(3) + scaleX(3), 
                                 scaleY(2),
                                 width - scaleX(2) - scaleX(3) - scaleX(3) - scaleX(2),  
                                 height - scaleY(2) - scaleY(2) - scaleY(5),
                                 false,
                                 faceColor);
                     
                     
                     // draw the lower window icon
                     
                     drawMaxRect(scaleX(3), 
                                 scaleY(3) + scaleY(3), 
                                 width - scaleX(2) - scaleX(3) - scaleX(3) - scaleX(2), 
                                 height - scaleY(2) - scaleY(2) - scaleY(5),
                                 true,
                                 faceColor);
                  }
                  break;
            }
         }
      });
   }

   /**
    * Set the button's dimension.
    * 
    * @param    width
    *           The button's width, in character units
    * @param    height
    *           The button's height, in character units
    */
   public void setDimension(double width, double height)
   {
      this.width = width;
      this.height = height;
   }
   
   /**
    * Get the button's {@link #width}.
    * 
    * @return   See above.
    */
   @Override
   public double width()
   {
      return width;
   }
   
   /**
    * Get the button's {@link #height}.
    * 
    * @return   See above.
    */
   @Override
   public double height()
   {
      return height;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      WindowGuiImpl window = (WindowGuiImpl) ancestor();

      switch (type)
      {   
          case CLOSE:
             // no-op - 4GL doesn't allow closing the app from the close button, 
             // but the WINDOW-CLOSE event is raised
             window.close();
             break;
          case ICONIFY:
             window.iconify();
             break;
          case MAXIMIZE:
             window.maximize();
             break;
      }
      
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      this.up = false;
      repaint();
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      this.up = true;
      repaint();
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return false;
   }
   
   /**
    * Draw a maximum-button style rectangle.
    * 
    * @param    x
    *           The X origin.
    * @param    y
    *           The Y origin.
    * @param    width
    *           The rectangle width.
    * @param    height
    *           The rectangle height.
    * @param    clear
    *           If <code>true</code>, clear the area with <code>clearColor</code>.
    * @param    clearColor
    *           Color to clear with.
    */
   private void drawMaxRect(int      x,
                            int      y,
                            int      width,
                            int      height, 
                            boolean  clear,
                            ColorRgb clearColor)
   {
      if (clear)
      {
         gd.setColor(clearColor);
         gd.fillRect(x, y, width, height);  
         gd.setColor(gd.getSysColor(SysColor.COLOR_BTNTEXT));
      }
      
      int strokeSize = scaleX(1);
      gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, strokeSize);
      
      gd.drawRect(x, y, width, height);

      gd.setLineStroke(LineStroke.SPECIFIED_WIDTH, scaleX(2));
      gd.drawLine(x + 1, y + 1, x + width - strokeSize, y + 1);
      
      gd.setLineStroke(LineStroke.DEFAULT);
   }
   
   /**
    * Updates the horizontal and vertical drawing scales.
    * 
    * @param    size
    *           The button's native dimensions.
    */
   private void updateScale(NativeDimension size)
   {
      xScale = (double) size.width / BASE_WIDTH;
      yScale = (double) size.height / BASE_HEIGHT;
   }
   
   /**
    * Scales the supplied coordinate value horizontally.
    * 
    * @param    value
    *           The value to scale.
    * 
    * @return   The scaled value.
    */
   private int scaleX(int value)
   {
      return (int) Math.round(xScale * value);
   }
   
   /**
    * Scales the supplied coordinate value vertically.
    * 
    * @param    value
    *           The value to scale.
    * 
    * @return   The scaled value.
    */
   private int scaleY(int value)
   {
      return (int) Math.round(yScale * value);
   }
}