/*
** Module   : RectangleGuiImpl.java
** Abstract : GUI rectangle implementation
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141114 Initial version.
** 002 EVL 20141120 Fixed origin coords inside painting worker.  Changed double line shaded
**                  rectangle drawing approach.  Also when GROUP-BOX attribute is true the filling
**                  become ignored.
** 003 CA  20141205 Avoid unnecessary conversion, as the physical location/dimension can be 
**                  determined from existing APIs.
** 004 EVL 20150409 Turning on clipping tool.
** 005 HC  20150323 Removed business logic from config classes.
** 006 GES 20150425 Eliminated AWT dependencies.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Specific GUI implementation of RECTANGLE widget.
 */
public class RectangleGuiImpl
extends Rect<GuiOutputManager>
{
   /** Screen driver. */
   private GuiDriver gd;

   /** Output manager. */
   private GuiOutputManager screen;
   
   /** Coordinate converter. */
   private CoordinatesConversion cc;

   /** Color resolver. */
   private GuiColorResolver gc;
   
   /** Rounding diameter in pixels. */
   private static final int ROUND_DIAMETER = 8;
   
   /** Light shadow box color. */
   private ColorRgb lightShadow;

   /** Dark shadow box color. */
   private ColorRgb darkShadow;
   
   /**
    * This constructor creates a rectangle.
    * 
    * @param    id
    *           The widget's ID.
    * @param    owner
    *           The parent container, frame which will include this
    *           rectangle.
    */
   RectangleGuiImpl(WidgetId id, Container<GuiOutputManager> owner) 
   {
      super(id, owner);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public RectangleGuiImpl(WidgetId id, RectangleConfig cfg)
   {
      super(id, cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, RectangleConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // initialize the shadow box colors
      lightShadow = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow  = gd.getSysColor(SysColor.COLOR_3DSHADOW);
   }
   
   /**
    * Draw the rectangle. Called by this rectangle's parent container.
    */
   @Override
   public void draw()
   {
      // rectangle options need to be refreshed before every painting
      final RectangleConfig config = (RectangleConfig) config();
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_WINDOW, SysColor.COLOR_WINDOWTEXT);
      
      // determine colors to use
      gc.colors();
   
      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      
      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));

      // check the current UI theme in use
      final boolean isXpStyle = gd.getUiTheme().equals("xp");
      
      // border size, GROUP-BOX is involved, consider Windows XP style or classic
      //    GROUP-BOX == true and Windows XP style -> single line rounded
      //    GROUP-BOX == true and Windows Classic style -> double line "shaded"
      final int borderSize = (config.groupBox && isXpStyle) ? 1 :
                             (config.groupBox && !isXpStyle) ? 2 : (int) config.edgePixels;
                             
      // is rectangle rounded or not
      //    GROUP-BOX == true means ROUNDED value is ignored
      final boolean isRounded = (config.groupBox && isXpStyle) ||
                                (!config.groupBox && config.rounded);
      
      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);
            // draw rectangle border
            if (borderSize > 0)
            {
               if (config.groupBox && !isXpStyle)
               {
                  drawShadedDoubleLineBorder(po, d, lightShadow, darkShadow);
               }
               else
               {
                  drawRectangleBorder(po, d, borderSize, gc.fgColor, isRounded);
               }
            }
            
            // fill rectangle, GROUP-BOX disables filling
            if (config.filled && borderSize >= 0 && !config.groupBox)
            {
               fillRectangle(po, d, borderSize, gc.bgColor, isRounded);
            }
         }
         
      });
   }

   /**
    * Draw rectangle border.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    borderSize
    *           Border size in pixels.
    * @param    color
    *           Border color.
    * @param    isRounded
    *           Flag indicating if the border is rounded or not.
    */
   private void drawRectangleBorder(NativePoint     orig, 
                                    NativeDimension dim, 
                                    int             borderSize,
                                    ColorRgb        color,
                                    boolean         isRounded)
   {
      gd.setColor(color);
      // draw rectangle borders from outmost to innermost
      if (isRounded)
      {
         for (int i = 0; i < borderSize; i++)
         {
            gd.drawRoundRect(orig.x + i, orig.y + i, dim.width - 2*i,
                             dim.height - 2*i, ROUND_DIAMETER);
         }
      }
      else
      {
         for (int j = 0; j < borderSize; j++)
         {
            gd.drawRect(orig.x + j, orig.y + j, dim.width - 2*j, dim.height - 2*j);
         }
      }
   }

   /**
    * Draw shaded double line rectangle border.  This means we simulate the embossing by combining
    * dark and light lines with colors provided.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    colorLight
    *           The light component of the shade.
    * @param    colorDark
    *           The dark component of the shade.
    */
   private void drawShadedDoubleLineBorder(NativePoint     orig,
                                           NativeDimension dim,
                                           ColorRgb        colorLight,
                                           ColorRgb        colorDark)
   {
      // upper left corner
      int x = orig.x;
      int y = orig.y;
      // width and height
      int w = dim.width - 1;
      int h = dim.height - 1;

      // starting with dark part
      gd.setColor(colorDark);
      gd.drawRect(x, y, w, h);

      // now light part
      gd.setColor(colorLight);
      gd.drawRect(x + 1, y + 1, w, h);
   }
   
   /**
    * Fill rectangle.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    borderSize
    *           Border size characters.
    * @param    color
    *           Fill color.
    * @param    isRounded
    *           Flag indicating if the border is rounded or not.
    */
   private void fillRectangle(NativePoint     orig, 
                              NativeDimension dim, 
                              int             borderSize,
                              ColorRgb        color,
                              boolean         isRounded)
   {
      gd.setColor(color);
      // exclude border from filling area
      int rx = orig.x + borderSize;
      int ry = orig.y + borderSize;
      int rw = dim.width - borderSize;
      int rh = dim.height - borderSize;
      
      if (isRounded)
      {
         gd.fillRoundRect(rx, ry, rw, rh, ROUND_DIAMETER);
      }
      else
      {
         gd.fillRect(rx, ry, rw, rh);
      }
   }
}
