/*
** Module   : ScrollBarGuiButton.java
** Abstract : scroll bar button.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 MAG 20140917 Created initial version.
** 002 MAG 20141020 Added cliping rectangle.
** 003 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 004 GES 20150507 Added a comment to clarify that AWT events usage is intentional.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Orientation;
import com.goldencode.p2j.ui.client.ScrollBar.Position;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * A scroll bar button.
 */
@SuppressWarnings("unchecked")
class ScrollBarGuiButton
extends Button<GuiOutputManager>
{   
   /** Scroll bar container */
   ScrollBar<GuiOutputManager> owner;
   
   /** The graphics driver. */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Button position */
   private ScrollBar.Position position;
   
   /** Button state */
   boolean up = true;
      
   /**
    * This constructor creates a scroll bar button.
    * 
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container.
    * @param   position
    *          Button position within scroll bar.
    */
   ScrollBarGuiButton(WidgetId id, ScrollBar<GuiOutputManager> owner, ScrollBar.Position position)
   {
      super(id, owner, "");
      
      this.owner = owner;
      this.position = position;
      
      _setVisible(true);
      gd = (GuiDriver) OutputManager.getDriver();
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_3DDKSHADOW);
   }

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      return new NativeDimension(config().widthPixels, config().heightPixels);
   }

   /**
    * Draw this scroll bar button, depending on its type.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }

      // resolve colors
      gc.colors();
      
      // get button coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      // clipping rectangle
      final NativeRectangle clip = clipRectangle(p, d);
      
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // draw button
            gd.setColor(gc.bgColor);
            gd.fill3DRect(0, 0, d.width, d.height, up);

            // draw arrow
            if (owner.orientation() == Orientation.HORIZONTAL)
            {
               int g = d.width / 3;   
               if (position == Position.LEFT)
               {
                  int[] xPoints = {g,          d.width-g, d.width-g};
                  int[] yPoints = {d.height/2, g,         d.height-g};
                  gd.setColor(gc.fgColor);
                  gd.fillPolygon(xPoints, yPoints);            
               }
               else
               {
                  int[] xPoints = {g, g,          d.width-g};
                  int[] yPoints = {g, d.height-g, d.height/2};
                  gd.setColor(gc.fgColor);
                  gd.fillPolygon(xPoints, yPoints);
               }               
            }   
            else
            {
               int g = d.height / 3;
               if (position == Position.LEFT)
               {
                  int[] xPoints = {d.width/2, g,          d.width-g};
                  int[] yPoints = {g,         d.height-g, d.height-g};
                  gd.setColor(gc.fgColor);
                  gd.fillPolygon(xPoints, yPoints);
               }
               else
               {
                  int[] xPoints = {g, d.width-g, d.width/2};
                  int[] yPoints = {g, g,         d.height-g};
                  gd.setColor(gc.fgColor);
                  gd.fillPolygon(xPoints, yPoints);
               }
            }
         }
      });
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */   
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // call scroll bar
      owner.onButtonClick(position);
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      up = false;
      redraw();         
   }

   /**
    * Notification of a mouse released event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      up = true;
      redraw();         
   }
   
   /**
    * Invoke repaint for this widget.
    */
   private void redraw()
   {
      ThinClient.getInstance().eventDrawingBracket(this, false, true, new Runnable()
      {
         @Override
         public void run()
         {
            repaint();
         }
      });      
   }   
}
