/*
** Module   : ScrollableListGuiImpl.java
** Abstract : scrollable list implementation. 
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141016 Improve text drawing and color management. Added mouse wheel handler. 
** 003 CA  20141123 Added scroll support via cursor keys.
** 004 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 005 HC  20150323 Removed business logic from config classes.
** 006 CA  20150507 Draw text centered vertically.
** 007 GES 20150425 Eliminated AWT dependencies.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Position;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.model.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Scrollable list GUI implementation. 
 */
public class ScrollableListGuiImpl
extends DefaultList<String, GuiOutputManager>
implements WidgetWithConfig<BaseConfig>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Widget configuration */
   private BaseConfig config;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Row height */
   private int rowHeight;

   /** The vertical scroll bar associated with this scrollable list. */
   private ScrollBar<GuiOutputManager> vbar;

   /** Flag indicating if the list elements can be unselected. */
   private final boolean allowDeselect;

   /**
    * Constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   model
    *          List model.
    * @param   allowDeselect
    *          <code>true</code> if the list elements can be unselected.
    */
   public ScrollableListGuiImpl(WidgetId id, ListModel<String> model, boolean allowDeselect)
   {
      super(model);
      this.allowDeselect = allowDeselect;
      gd = (GuiDriver) OutputManager.getDriver();
      config = ConfigManager.getInstance().resolveWidgetConfig(id, BaseConfig.class);
      config.visible = true;
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);
      gc.selectionBgColor = gd.getSysColor(SysColor.COLOR_HOTLIGHT);
      gc.selectionFgColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
      // margins
      nativeInsets = new NativeInsets(0, 1, 0, 1);
   } 
         
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, BaseConfig cfg)
   {
      // no-op
   }

   /**
    * Process scroll event.
    * 
    * @param   event
    *          Scroll event.
    */
   @Override
   public void processScrollEvent(ScrollEvent event)
   {
      super.ensureRowVisible((int) event.limit().y, false);       
      repaint();
   }

   /**
    * Get the widget config.
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   @Override
   public BaseConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(BaseConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      boolean parentWidth = (isInheritWidth() && parent() != null);
      return (parentWidth) ? parent().width() : super.width();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      boolean parentHeight = (isInheritHeight() && parent() != null);
      return (parentHeight) ? parent().height() : super.height();
   }
   
   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible
      if (!isVisible())
      {
         return;
      }
      
      // resolve font
      final FontCache cached = font();
      
      // resolve colors
      gc.colors();
      
      // get container coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = clipRectangle(p, d);

      // display rows
      final int row  = this.currentRow();
      final int rows = this.visibleRows();
      
      // model size
      final int modelSize = model().size();

      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // draw list
            gd.setColor(gc.bgColor);
            gd.fillRect(nativeInsets.left, nativeInsets.top, d.width, d.height);
            
            if (modelSize > 0 && cached.font != null)
            {
               gd.setGuiFont(cached.font);
               gd.setFontStyle(cached.font.style);

               for (int i = row, l = 0; i < (row + rows); i++, l++)
               { 
                  if (i >= 0 && i < modelSize)
                  {
                     int x = nativeInsets.left;
                     int y = rowHeight * l;
                     int dx = d.width;
                     int dy = rowHeight;
                     
                     if (selected(i))
                     {
                        gd.setColor(gc.selectionBgColor);
                        gd.fillRect(x, y, dx, dy);
                        gd.setColor(gc.selectionFgColor);
                     }
                     else
                     {
                        gd.setColor(gc.bgColor);
                        gd.fillRect(x, y, dx, dy);
                        gd.setColor(gc.fgColor);
                     }
                     
                     // draw item text
                     gd.translate(x, y);
                     String text = model().get(i);
                     if (text != null)
                     {
                        gd.drawStringCentered(text, nativeInsets.left, rowHeight - nativeInsets.top);
                     }
                     gd.translate(-x, -y);
                  }
               }
            }
         }
      });
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      synchronized (this)
      {
         if (model().size() > 0 && rowHeight > 0)
         {
            NativePoint n = translate(this, e);
            
            // calculate clicked line
            int row = n.y / rowHeight + currentRow();
            
            if (selected(row))
            {
               if (allowDeselect && selectionMode() == ListSelectionModel.SINGLE_SELECTION)
               {
                  deselect(row, row);
               }
            }
            else
            {
               select(row);
            }
            
            repaint(); 
         }
      }

      ThinClient.getInstance().getFocusManager().setFocusOn(this);
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // total items in the list
      int items = model().size(); 
      
      if (items > 0)
      {
         int item = currentRow() + e.getWheelRotation();
         int max = items - visibleRows();

         if (item >= max )
         {
            item = max;
         }
         else if (item < 0)
         {
            item = 0;
         }

         // scroll list
         ensureRowVisible(item);
         // notify listeners
         super.processScrollEvent(new ScrollEvent(this, ScrollDirection.UP, new Point(0, item)));         
         repaint();
      }      
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      int[] selection = selection();
      
      if (selection.length == 0)
      {
         // nothing to do, go with default
         super.processKeyEvent(event);
         return;
      }

      if (selection.length > 1)
      {
         UnimplementedFeature.partial(
            "Key received for a scrollable list with multiple selected elements");
      }

      int item = selection[0];
      int oldRow = currentRow();
      Position pos = null;
      
      switch (event.keyCode())
      {
         case Key.VK_CURSOR_UP:
            item = item - 1;
            pos = Position.LEFT;
            break;
         
         case Key.VK_CURSOR_DOWN:
            item = item + 1;
            pos = Position.RIGHT;
            break;
            
         default:
            // for other keys, go with default
            super.processKeyEvent(event);
            return;
      }
      
      if (item < 0 || item > model().size() - 1)
      {
         // reached top/bottom limit
         event.consume();
         return;
      }
      
      select(item);
      
      if (vbar != null && (item < oldRow || item > oldRow + visibleRows() - 1))
      {
         // simulate a knob drag so that the currently selected item is brought to the visible
         // area
         int current = item < oldRow ? item : item - visibleRows() + 1;
         vbar.onDrag(pos, current);
      }
      
      repaint();
   }

   /**
    * Add new item to list.
    * 
    * @param   item
    *          Item to add.
    *          
    * @return  First visible row. Could be negative.         
    */
   public int addItem(String item)
   {
      model().add(item);
      int firstVisibleRow = model().size() - visibleRows();
      ensureRowVisible(firstVisibleRow);
      return firstVisibleRow;
   }
   
   /**
    * Delete all items.
    */
   public void deleteItems()
   {
      model().clear();
      ensureRowVisible(0);
   }

   /**
    * Set the list's vertical scrollarr.
    * 
    * @param    vbar
    *           The vertical scrollbar.
    */
   public void setScrollBar(ScrollBar<GuiOutputManager> vbar)
   {
      this.vbar = vbar;
   }

   /**
    * Resolve font. Because the font could be changed dynamically whenever we draw the widget we
    * have to check if font was changed and rebuild font object if true.
    */
   private FontCache font()
   {
      if (gf == null)
      {
         Window<?> window = (Window<?>) ancestor();
         gf = new GuiFontResolver(window, gd, config, false);      
      }
      
      FontCache cached = gf.font();
      rowHeight = cached.fontHeight + nativeInsets.top + nativeInsets.bottom;
      return cached;
   }
}
