/*
** Module   : ToggleBoxGuiImpl.java
** Abstract : GUI toggle-box implementation
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 VIG 20150210 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
*/
package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.Container;

/**
 * Specific GUI implementation of TOGGLE-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ToggleBoxGuiImpl
extends ToggleBox<GuiOutputManager>
{
   /** The space between checkbox and label */
   private static final int SPACE = 6;

   /** Checkbox width or length */
   private static int CHKBOX_WIDTH = 11;
   
   /** Color resolver */
   private GuiColorResolver gc;

   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver */
   private GuiFontResolver gf;
   
   /** Checkbox shadow color. */
   private ColorRgb chkBoxLTShadow;
   
   /** Checkbox left top sides color. */
   private ColorRgb chkBoxLTColor;
   
   /** Checkbox right bottom shadow color. */
   private ColorRgb chkBoxRBShadow;
   
   /** Checkbox right bottom sides color. */
   private ColorRgb chkBoxRBColor;
   
   /** Check box inactive background color. */
   private ColorRgb chkBoxBGInactive;
   
   /** Check mark color. */
   private ColorRgb chkMarkColor;
   
   /** Label inactive background color. */
   private ColorRgb labelInactiveColor;
                      
   /**
    * Constructor.
    *
    * @param   id
    *          The widget's ID.
    * @param   owner
    *          The parent container, frame or dialog which will include this toggle-box.
    * @param   text
    *          The string to be used as the toggle-box text.
    */
   public ToggleBoxGuiImpl(WidgetId id, Container<GuiOutputManager> owner, String text)
   {
      super(id, owner, text);
   }

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           Configuration data.
    */
   public ToggleBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ToggleBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ToggleBoxConfig wcfg)
   {
      super.initialize(id, wcfg);

      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();

      // font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, config());

      // initialize the colors
      chkBoxLTShadow     = gd.getSysColor(SysColor.COLOR_3DDKSHADOW); 
      chkBoxLTColor      = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      chkBoxRBShadow     = gd.getSysColor(SysColor.COLOR_BTNHIGHLIGHT);
      chkBoxRBColor      = gd.getSysColor(SysColor.COLOR_3DLIGHT);
      chkBoxBGInactive   = gd.getSysColor(SysColor.COLOR_MENU);
      chkMarkColor       = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
      
      // TODO find out what color or font of disabled label 
      labelInactiveColor = gd.getSysColor(SysColor.COLOR_3DFACE);
      
      // register tooltip action listener
      gd.registerTooltipWorker(this, gf.font().font, ToolTip.isActive(config().tooltip));
      
      if (config.autoResize)
      {
         setWidthInt(); 
         config.heightChars = 0.81;
      }
   }

   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    *           Previous configuration values.
    */
   @Override
   public void afterConfigUpdate(ToggleBoxConfig beforeUpdate)
   {
      if (beforeUpdate.label != config().label && config.autoResize)
      {
         setWidthInt();

         // TODO figure out why SyncCoordinatesAspect is not called here and use it.
         CoordinatesConversion cc = screen().coordinates();
         config.widthPixels = cc.columnToPixels(config.widthChars);
         config.widthChars = cc.columnFromPixels(config.widthPixels); 
      }
      
      super.afterConfigUpdateBase(beforeUpdate);
   }
   
   /**
    * Draw the toggle-box implementation. 
    */
   @Override
   public void draw()
   {
      gc = new GuiColorResolver(gd, config, SysColor.COLOR_FILLIN_BG, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      final String text = config.label;
      final FontCache fntCache = gf.font();
      
      // get location and size
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width + 1, d.height + 1));
      
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // clear the area
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);
            
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            // toggle-box state flags
            boolean highlighted = highlighted();
            
            // focused toggle-box needs special highlight
            if (highlighted)
            {
               drawHighlight(po, d, text, gc.fgColor, fntCache.font);
            }

            drawCheckBox(po, d, fntCache.font);

            int xt = po.x + CHKBOX_WIDTH + SPACE; // shift between checkbox and label
            int yt = po.y + (d.height + fntCache.font.pointSize)/2;  
            
            gd.setColor(isEnabled() ? gc.fgColor : labelInactiveColor);
            gd.setGuiFont(fntCache.font);
            gd.setFontStyle(fntCache.font.style);
            gd.drawString(text, xt, yt);
         }
      });
   }

   /**
    * Recalculates WIDTH attribute using actual label.
    */
   public void setWidthInt()
   {
      Window<?> window = Window.resolveWindow(this);
      // method calls server-side to get legacy text metrics, so current changes could be sent.
      int nativeWidth = FontManager.getTextWidth(config.label, window, gf.resolveFontNum(), gd);
      CoordinatesConversion cc = screen().coordinates();
      config.widthChars = cc.widthFromNative(nativeWidth) + 5;
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.  Simulate ENTER key typing.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // simulate enter key
      if (isEnabled())
      {
         check();
         e.consume(); // ?
         this.requestFocus();
      }
   }
   
   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // TODO implement disabling during dragging
   }
   
   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // TODO implement disabling during dragging
   }

   /**
    * In GUI mode, the drawing needs to be delayed.
    */
   @Override
   protected void refresh()
   {
      repaint();
   }
   
   /**
    * Draw highlight frame indicating focused button.
    * 
    * @param    orig
    *           Rectangle origin.
    * @param    dim
    *           Rectangle size.
    * @param    color
    *           Border color.
    */
   private void drawHighlight(NativePoint     orig,
                              NativeDimension dim,
                              String          label,
                              ColorRgb        color,
                              FontDetails<?>  font)
   {
      gd.setColor(color);
   
      // set dotted line style      
      gd.setLineStroke(LineStroke.DOTTED_SMALL_THIN);
      
      int textWidth = gd.getTextWidth(label, font);
      int textHeight = gd.getTextHeight(label, font);
            
      int xt = orig.x + CHKBOX_WIDTH + SPACE;
      int yt = orig.y - 1 + (dim.height - textHeight) / 2; 
      
      gd.drawRect(xt, yt, textWidth, textHeight + 3);
      
      // restore default line style
      gd.setLineStroke(LineStroke.DEFAULT);
   }

   /**
    * Draws check box 
    * 
    * @param    po
    *           origin point.
    * @param    d
    *           Native toggle-box dimension.
    * @param    font
    *           Configured font.       
    */
   private void drawCheckBox(NativePoint po, NativeDimension d, FontDetails<?> font)
   {
      int w = CHKBOX_WIDTH;
      NativePoint p = new NativePoint(po.x, po.y + (d.height - w) / 2);
      
      // draw background
      gd.setColor(isEnabled() ? chkBoxRBShadow : chkBoxBGInactive);
      gd.fillRect(p.x, p.y, w + 1, w + 1);
      
      // left  top angle
      gd.setColor(chkBoxLTColor);
      gd.drawLine(p.x, p.y, p.x + w, p.y);
      gd.drawLine(p.x, p.y + 1, p.x, p.y + w);
      gd.setColor(chkBoxLTShadow);
      gd.drawLine(p.x + 1, p.y + 1, p.x + w - 1, p.y + 1);
      gd.drawLine(p.x + 1, p.y + 1, p.x + 1, p.y + w - 1);
      
      // right bottom angle
      gd.setColor(chkBoxRBColor);
      gd.drawLine(p.x + 2, p.y + w, p.x + w, p.y + w);
      gd.drawLine(p.x + w, p.y + 2, p.x + w, p.y + w);
      gd.setColor(chkBoxRBShadow); 
      gd.drawLine(p.x, p.y + w + 1, p.x + w + 1, p.y + w + 1);
      gd.drawLine(p.x + w + 1, p.y, p.x + w + 1, p.y + w);
      
      // draw check
      if (config.checked)
      {
         gd.setColor(chkMarkColor);
         // left half
         gd.drawLine(p.x + 3, p.y + 5, p.x + 5, p.y + 7);
         gd.drawLine(p.x + 3, p.y + 6, p.x + 5, p.y + 8);
         gd.drawLine(p.x + 3, p.y + 7, p.x + 5, p.y + 9);
         // right half
         gd.drawLine(p.x + 9, p.y + 3, p.x + 5, p.y + 7);
         gd.drawLine(p.x + 9, p.y + 4, p.x + 5, p.y + 8);      
         gd.drawLine(p.x + 9, p.y + 5, p.x + 5, p.y + 9);      
      }
   }
   
}
