/*
** Module   : WindowLayout.java
** Abstract : Layout the elements of a window.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140912 Set a new BitmapScreen object when window is resized.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140923 GUI message area layout.
** 006 CA  20141021 Fixes related to window layout during resize.
** 007 MAG 20141021 Clear window borders whenever is resized.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed initial dimension for dynamic windows.
** 011 HC  20141207 Improved window layout algorithm to honor the requested window size.
** 012 HC  20150323 Removed business logic from config classes.
** 013 CA  20150507 Fixed problems related to window sizing (explicit or implicit).
** 014 CA  20150508 Fixed H013 regression when resizing window via mouse (not attributes).
** 015 HC  20150508 Improvements to GUI window scrolling support.
** 016 GES 20150425 Match class name change for GuiOptions to GuiConstants.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Layout the elements of a legacy-style window. This includes the window title, the workspace,
 * message area and status area.
 */
public class WindowLayout
implements LayoutManager<GuiOutputManager>
{
   /**
    * Get the minimum size of the specified container.
    * 
    * @param    container
    *           The container.
    *          
    * @return   Always {@code null}.
    */
   @Override
   public Dimension minimumSize(Container<GuiOutputManager> container)
   {
      return null;
   }

   /**
    * Place components in specified container.
    * 
    * @param   container
    *          Container to process - must be a window.
    */
   @Override
   public void doLayout(Container<GuiOutputManager> container)
   {
      // TODO: only use pixel/native coordinates, syncing of the characters values
      //       must be handled by the framework
      
      WindowGuiImpl window = (WindowGuiImpl) container;
      
      CoordinatesConversion c = window.screen().coordinates();

      // select renderer
      window.gd.selectWindow(window.config().id.asInt());

      // 0. prerequisites for computing full width/height:
      boolean resizing = window.isResizing();

      // - the message area height
      int mheight = c.heightToNative(window.getGuiMessageArea().height());
      
      // - the status area height
      int sheight = c.heightToNative(window.getStatusArea().height()); 

      // - the border delta
      int borderDelta = GuiConstants.BORDER_SIZE * 2;
      
      // - window size
      final int    wheight;
      final int    wwidth;
      final double cwheight;
      final double cwwidth;
      // - workspace size
      final double cheight;
      final double cwidth;
      int height;
      int width;
      
      WindowConfig wc = window.config();
      ScrollPaneConfig wwc = window.workspace.config();
      
      if (resizing)
      {
         NativeDimension oldSize = window.physicalDimension();
         
         // - get the new workspace height/width
         width = wwc.widthPixels + (window.resizedWidth - oldSize.width);
         height = wwc.heightPixels + (window.resizedHeight - oldSize.height);
         
         int minmaxWidthDelta = 0;
         int minmaxHeightDelta = 0;
         
         // limit the new width by the config min/max width
         if (width < wc.minWidthPixels)
         {
            minmaxWidthDelta = wc.minWidthPixels - width;
            width = wc.minWidthPixels;
         }
         else if (width > wc.maxWidthPixels)
         {
            minmaxWidthDelta = wc.maxWidthPixels - width;
            width = wc.maxWidthPixels;
         }
         
         // limit the new height by the config min height
         if (height < wc.minHeightPixels)
         {
            minmaxHeightDelta = wc.minHeightPixels - height;
            height = wc.minHeightPixels;
         }
         else if (height > wc.maxHeightPixels)
         {
            minmaxHeightDelta = wc.maxHeightPixels - height;
            height = wc.maxHeightPixels;
         }
         
         cheight = c.heightFromNative(height);
         cwidth  = c.widthFromNative(width); 
         
         wheight = window.resizedHeight + minmaxHeightDelta;
         wwidth = window.resizedWidth + minmaxWidthDelta;
         cwheight = c.heightFromNative(wheight);
         cwwidth = c.widthFromNative(wwidth);

         // - set the titlebar's width
         double ctbwidth = c.widthFromNative(wwidth - borderDelta);
         window.titleBar.setWidth(ctbwidth);
         window.titleBar.doLayout();
      }
      else
      {
         // this is the workspace size.  it is what the 4GL HEIGHT-/WIDTH- attributes are referring
         cheight = wc.heightChars;
         cwidth  = wc.widthChars;
         height  = c.heightToNative(cheight);
         width   = c.widthToNative(cwidth);

         // - layout the titlebar now, is needed to compute the window height
         double ctbwidth = c.widthFromNative(width);
         window.titleBar.setWidth(ctbwidth);
         window.titleBar.doLayout();
         int tbheight = c.heightToNative(window.titleBar.height());

         wheight  = height + tbheight + mheight + sheight + borderDelta; 
         wwidth   = width + borderDelta;
         cwheight = c.heightFromNative(wheight);
         cwwidth  = c.widthFromNative(wwidth); 
      }
      
      int tbheight = c.heightToNative(window.titleBar.height());
      
      // TODO: in addition to MIN/MAX-HEIGHT attrs the window seems to have
      //       physical size limits - for width it is a minimum window title bar plus borders
      //       and for height it is the title bar height plus message area plus window status bar
      //       plus borders.
      
      window.workspace.setSize(cwidth, cheight);
      wc.widthChars = cwidth;
      wc.heightChars = cheight;
      wc.widthPixels = width;
      wc.heightPixels = height;

      wwc.widthChars = cwidth;
      wwc.heightChars = cheight;
      wwc.widthPixels = width;
      wwc.heightPixels = height;
      
      if (resizing)
      {
         // clear window border
         clearBorder(window.gd, new NativeDimension(wwidth, wheight));
      }
      else
      {
         window.gd.resizeWindow(wwidth, wheight);
      }

      window.setSize(cwwidth, cwheight);
      
      // update screen bitmap
      window.setBitmap(new ScreenBitmap(wheight, wwidth));
      
      // all coordinates for the following widgets are relative to the OS window (the container)
      int x = GuiConstants.BORDER_SIZE;
      int y = GuiConstants.BORDER_SIZE;
      
      // 1. set the title into position:
      window.titleBar.setPhysicalLocation(x, y);
      window.titleBar.setLocation(c.widthFromNative(x), c.heightFromNative(y));
      
      // 2. set the window workarea (this is the legacy 4GL window, where frames are placed)
      // - bellow the title and above the message area
      // - its size does not include the borders
      // - its size needs to be computed first, as the general window size depends on it
      y = y + tbheight;
      window.workspace.setPhysicalLocation(x, y);
      window.workspace.setLocation(c.widthFromNative(x), c.heightFromNative(y)); 
      
      // 3. set the message area
      y = y + height;
      window.getGuiMessageArea().setPhysicalLocation(x, y);
      window.getGuiMessageArea().setLocation(c.widthFromNative(x), c.heightFromNative(y)); 
      window.getGuiMessageArea().doLayout();
      
      // 4. set the status area
      y = y + mheight;
      window.getStatusArea().setPhysicalLocation(x, y);
      window.getStatusArea().setLocation(c.widthFromNative(x), c.heightFromNative(y));
   }

   /**
    * Whenever window is resized the right and bottom borders became dirty.  In order to prevent
    * dirty borders right and bottom borders are cleaned.
    * 
    * @param   gd
    *          GUI device.
    * @param   d
    *          Window dimension.
    */
   private void clearBorder(GuiDriver gd, NativeDimension d)
   {
      // main window border size
      int border = GuiConstants.BORDER_SIZE;
      
      // clear left and bottom borders
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.fillRect(d.width - border, 0, border, d.height);
      gd.fillRect(0, d.height - border, d.width, border);      
   }
}
