/*
** Module   : p2j.screen.js
** Abstract : GUI-specific screen management
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 GES 20150325 Created GUI version from ChUI version. Right now it is not functional.
**                  Renamed objects to match the more generic web client approach (not just
**                  ChUI). Moved some logic to common code locations.
*/

p2j.screen = (function()
{
   /** Screen object which will be returned from this initialization function. */
   var me = {};

   /** Main drawing canvas. */
   var canvas;

   /** Canvas 2D graphics context. */
   var ctx;

   /** Normal canvas font. */
   var font, bold;

   /** Bold canvas font. */
   var bold;
   
   /** Rows on the screen. */
   var rows;

   /** Columns on the screen. */
   var cols;
   
   /** Font width. */
   var dx;

   /** Font height. */
   var dy;

   /** Color palette. */
   me.palette;

   /**
    * Module   : video
    * Abstract : screen video memory
    */
   var video = (function()
   {
      /** Video memory object which will be returned from this initialization function. */
      var me = {};

      /** Internal video storage. */
      var memo;

      /**
       * Initialize video memory.
       * 
       * @param {number} rows    total number of rows
       * @param {number} cols    total number of columns
       * @param {number} color   color pair FG/BG
       */
      me.init = function(rows, cols, color)
      {
         me.color = color;
         memo = new Array(rows);
         for (i = 0; i < rows; i++)
         {
            memo[i] = new Array(cols);
            for (j = 0; j < cols; j++)
            {
               var cell = {'c': ' ', 'f': 'white', 'b': 'black', 'a': 0, 'k': 0};
               cell.f = color.f;
               cell.b = color.b;
               memo[i][j] = cell;
            }
         }
      };

      /**
       * Clear video memory.
       * 
       * @param {number} rows total number of rows
       * @param {number} cols total number of columns
       */
      me.clear = function(rows, cols)
      {
         for (i = 0; i < rows; i++)
         {
            for (j = 0; j < cols; j++)
            {
               var v = memo[i][j];
               v.c = ' ';
               v.a = 0;
               v.k = 0;
               v.f = me.color.f;
               v.b = me.color.b;
            }
         }
      };

      /**
       * Get cell value.
       * 
       * @param {number} row  row number
       * @param {number} col  column number
       * 
       * @returns {object} cell value.
       */
      me.get = function(row, col)
      {
         return memo[row][col];
      };

      /**
       * Set cell value.
       * 
       * @param {char}   c   character
       * @param {number} row row 
       * @param {number} col column
       * @param {number} a   attributes
       * @param {number} k   color
       * @param {object} p   color pair FG/BG
       */
      me.put = function(c, row, col, a, k, p)
      {
         var v = memo[row][col];
         v.c = c;
         v.a = a;
         v.k = k;
         v.f = p.f;
         v.b = p.b;
      };

      /**
       * Copy selected area into clipboard.
       * 
       * @param {number} begRow start row
       * @param {number} endRow end row
       * @param {number} begCol start column
       * @param {number} endCol end column
       * 
       * @returns {object} selected text.
       */
      me.copy = function(begRow, endRow, begCol, endCol)
      {
         var text = "";         
         for (i = begRow; i <= endRow; i++)
         {
            for (j = begCol; j <= endCol; j++)
            {
               text += memo[i][j].c;
            }
         }         
         return text;
      };

      return me;
   })();

   /**
    * Module   : cursor 
    * Abstract : screen cursor management
    */
   var cursor = (function(screen)
   {
      /** Cursor object which will be returned from this initialization function. */
      var me = {};

      /**
       * Hide cursor.
       */
      me.hide = function()
      {
         screen.drawCursor(false);
      };

      /**
       * Show cursor.
       */
      me.show = function()
      {
         screen.drawCursor(me.visible);
      };

      /**
       * Set cursor position.
       * 
       * @param {number} row row number
       * @param {number} col column number
       */
      me.setPosition = function(row, col)
      {
         me.hide();
         me.row = row;
         me.col = col;
         me.show();
      };

      /**
       * Module   : blinking 
       * Abstract : screen cursor blinking management
       */
      me.blinking = (function(cursor, screen)
      {
         /** Cursor blink object which will be returned from this initialization function. */
         var me = {};

         /** timer */
         var timer;

         /**
          * Blinking off.
          */
         me.off = function()
         {
            window.clearInterval(timer);
         };

         /**
          * Blinking on.
          */
         me.on = function()
         {
            var blink = true;
            timer = window.setInterval(function()
            {
               screen.drawCursor(blink & cursor.visible);
               blink = !blink;
            }, 500);
         };

         return me;
      }(me, screen));

      /**
       * Initialize cursor.
       * 
       * @param {object} cursor attributes.
       */
      me.init = function(cursor)
      {
         me.row = 0;
         me.col = 0;
         me.type = cursor.type;
         me.visible = cursor.visible;
         if (cursor.blinking)
         {
            me.blinking.on();
         }
      };

      return me;
   })(me);

   /**
    * Draw cursor.
    * 
    * @param {boolean} on  true  = show cursor
    *                      false = hide cursor
    */
   me.drawCursor = function(on)
   {
      var v = video.get(cursor.row, cursor.col);
      var p = getColors(v.k);
      var x = cursor.col * dx;
      var y = cursor.row * dy;

      if (on)
      {
         if (cursor.type === 'solid')
         {
            p.f = video.color.b;
            p.b = video.color.f;
            drawText(x, y, v.c, 0, p);
         }
         else
         {
            drawText(x, y, v.c, v.a, p);
            // draw line cursor           
            ctx.fillStyle = video.color.f;
            ctx.fillRect(x, y + dy - 4, dx, 4);
         }
      }
      else
      {
         drawText(x, y, v.c, v.a, p);         
      }      
   };

   /**
    * Draw text on canvas.
    * 
    * @param {number} x      x coordinate
    * @param {number} y      y coordinate
    * @param {string} text   text to draw
    * @param {number} a      attribute
    * @param {object} v      color pairs
    */
   var drawText = function(x, y, text, a, v)
   {
      var max = dx * text.length;

      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y, max, dy);
      ctx.clip();

      ctx.font = (a & 0x02) ? bold : font;
      ctx.textBaseline = 'top';

      /** reverse */
      if (a & 0x08)
      {
         ctx.fillStyle = v.f;
         ctx.fillRect(x, y, max, dy);
         ctx.fillStyle = v.b;
      }
      else
      {
         ctx.fillStyle = v.b;
         ctx.fillRect(x, y, max, dy);
         ctx.fillStyle = v.f;
      }
      
      ctx.fillText(text, x, y, max);

      // underline
      if (a & 0x04)
      {
         ctx.fillRect(x, y + dy - 1, max, 1);
      }

      ctx.restore();                  
   };

   /**
    * Print an error message on the last line.
    * 
    * @param {string} text   text to print.
    */
   me.error = function(text)
   {
      var y = (rows - 1) * dy;

      var v = {};
      v.f = video.color.f;
      v.b = video.color.b;

      drawText(0, y, text, 0, v);

      p2j.sound.beep();
   };

   /**
    * Initialize module.
    *  
    * @param {object} cfg configuration.
    */
   me.init = function(cfg)
   {
      if (cfg.font.name === undefined || cfg.font.name.toLowerCase() === 'monospaced')
      {
         cfg.font.name = 'monospace';
      }
      
      rows = cfg.canvas.size.rows;
      cols = cfg.canvas.size.cols;
      
      video.init(rows, cols, cfg.font.color);
      
      canvas = document.getElementById(cfg.canvas.id);
      ctx = canvas.getContext('2d');
      
      font = cfg.font.size + 'px ' + cfg.font.name;
      bold = 'bold ' + font;
      
      ctx.font = font;
      
      dx = ctx.measureText('\u2500').width;
      dy = fontMetrics.cellHeight(font);
      
      canvas.width = cols * dx;
      canvas.height = rows * dy;
      
      ctx.fillStyle = cfg.font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      
      cursor.init(cfg.cursor);
      
      // TODO: was it OK to move the key event listener registration from here to p2j.keyboard.js?
      
      if (cfg.canvas.mouse)
      {
         // mouse click
         canvas.addEventListener('click', function(evt)
         {
            var rect = canvas.getBoundingClientRect();
            var x = evt.clientX - rect.left;
            var y = evt.clientY - rect.top;
            var c = (x / dx) | 0;
            var r = (y / dy) | 0;
            if ((r >= 0 && r < rows) && (c >= 0 && c < cols))
            {
               cursor.setPosition(r, c);
            }
         }, true);
      }
      
      // export some values
      me.canvasWidth = canvas.width;
      me.canvasHeight = canvas.height;
      me.charWidth = dx;
      me.charHeight = dy;
      
      // container center aligned
      document.getElementById(cfg.container).style.width = me.canvasWidth + 'px';
      
      canvas.focus();
   };

   /**
    * Set cursor position.
    * 
    * @param {number} row   row number
    * @param {number} col   column number
    */
   me.setCursorPosition = function(row, col)
   {
      cursor.setPosition(row, col);
   };

   /**
    * Set cursor visibility
    * 
    * @param {boolean} visible   true  = visible
    *                            false = hidden
    */
   me.setCursorStatus = function(visible)
   {
      cursor.visible = visible;
      cursor.show();
   };

   /**
    * Clear the screen.
    */
   me.clear = function()
   {
      video.clear(rows, cols);
      ctx.fillStyle = font.color.b;
      ctx.fillRect(0, 0, canvas.width, canvas.height);
      cursor.setPosition(0, 0);
   };

   /**
    * Get color pair from cell color.
    *  
    * @param {number} k    cell color
    * 
    * @returns {object} color pair.
    */
   var getColors = function(k)
   {
      var p = {};

      if (k === 0)
      {
         p.f = video.color.f;
         p.b = video.color.b;
      }
      else
      {
         // check for colors
         var fg = ((k >>> 4) & 0x0f) - 1;
         var bg = (k & 0x0f) - 1;

         p.f = me.palette[fg];
         p.b = me.palette[bg];
      }

      return p;
   };

   /**
    * Draw rectangles.
    *  
    * @param {array} rects  an array of rectangles
    */
   me.drawRectangles = function(rects)
   {
      var off = 1;

      while (off < rects.length)
      {
         // package length
         var len = (rects[off] << 8) | rects[off+1];
         me.drawRectangle(rects.subarray(off, off + len));
         off += len;
      }
   };

   /**
    * Draw a rectangle.
    * 
    * @param {array} rect   rectangle to draw
    */
   me.drawRectangle = function(rect)   
   {
      var off = 8; // header size
      // 0 and 1 are used for packege length. 
      var w = rect[2]; // width
      var h = rect[3]; // height
      var row = rect[4]; // row
      var col = rect[5]; // col
      var y = row * dy;
      var x = col * dx;
      var a = rect[6]; // attributes
      var k = rect[7]; // color
      var rx = w * dx; // rectangle width in pixels
      var ry = h * dy; // rectangle height in pixels
      var ls = w * 2;  // text line size

      // colors     
      var p = getColors(k);

      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y, rx, ry);
      ctx.clip();

      ctx.fillStyle = (a & 0x08) ? p.f : p.b;
      ctx.fillRect(x, y, rx, ry);
      // prepare for text draw
      ctx.font = (a & 0x02) ? bold : font;
      ctx.fillStyle = (a & 0x08) ? p.b : p.f;
      ctx.textBaseline = 'top';
      
      // draw text line by line
      for (var i = 0;i < h; i++)
      {
         me.drawRectangleLine(row + i, col, rect, off, ls, a, k , p, y);
         off += ls;
      }            
      
      ctx.restore();                  
   };

   /**
    * Draw a line within rectangle.
    * 
    * @param {number} row    row
    * @param {number} col    column
    * @param {array}  line   text
    * @param {number} off    offset
    * @param {number} len    line size
    * @param {number} a      attributes
    * @param {number} k      color
    * @param {object} p      color pair FG/BG
    */
   me.drawRectangleLine = function(row, col, line, off, len, a, k, p)
   {
      var j = col;
      
      for (var i = off; i < off + len; i += 2)
      {
         var c = String.fromCharCode(line[i] << 8 | line[i + 1]);
         me.drawCell(c, row, j);
         video.put(c, row, j++, a, k, p);
      }      

      var y = row * dy;
      var x = col * dx;
      var size = dx * len / 2;

      // underline
      if (a & 0x04)
      {
         ctx.fillRect(x, y + dy - 1, size, 1);
      }      
   };

   /**
    * Draw character in a cell.
    * 
    * @param {tring}  text charcater to draw
    * @param {number} row  row number
    * @param {number} col  columns number
    */
   me.drawCell = function(text, row, col)
   {
      var y = row * dy;
      var x = col * dx;
      
      ctx.save();
      ctx.beginPath();
      ctx.rect(x, y, dx, dy);
      ctx.clip();

      ctx.fillText(text, x, y, dx);      
      
      ctx.restore();                  
   };
   
   /**
    * Module   : fontMetrics
    * Abstract : measure font height.
    */
   var fontMetrics = (function()
   {
      /** Font metrics object which will be returned from this initialization function. */
      var me = {};

      /** Canvas height. */
      var height = 128;
      
      /** Canvas width. */
      var width  = 128;

      /** Canvas use to measure metrics. */
      var fontCanvas;

      /** 2D graphical context. */
      var fontCtx;
      
      /**
       * Get font height.
       * 
       * @param {string} font
       * 
       * @returns {number} font heigth.
       */
      me.cellHeight = function(font)
      {
         fontCanvas = document.createElement('canvas');
         fontCanvas.setAttribute('height', height);
         fontCanvas.setAttribute('width', width);
         fontCtx = fontCanvas.getContext('2d');
         fontCtx.fillStyle = 'black';
         fontCtx.fillRect(0, 0, width, height);
         fontCtx.textBaseline = 'top';
         fontCtx.fillStyle = 'white';
         fontCtx.font = font;
         fontCtx.fillText('\u2503Eghy', 0, 0);

         var pixels = fontCtx.getImageData(0, 0, width, height).data;

         // row numbers where we first find letter end where it ends 
         var beg = -1;
         var end = -1;

         for (var row = 0; row < height; row++)
         {
            for (var col = 0; col < width; col++)
            {
               var index = (row * width + col) * 4;
               
               // if pixel is not white (background color)
               if (pixels[index] === 0)
               {
                  // we havent met white (font color) pixel
                  // on the row and the letters was detected
                  if (col === (width - 1) && beg !== -1)
                  {
                     end = row;
                     row = height;
                     break;
                  }
               }
               else
               {
                  // we find top of letter
                  if (beg === -1)
                  {
                     beg = row;
                  }
                  // ..letters body
                  break;
               }
            }
         }

         // space at top
         me.top = beg;
         
         // character height
         me.charHeight = end - beg;
         
         // cell height top + charHeight
         me.cellHeight = end;
         
         // cell height
         return p2j.isWindows ? end : end - 1;
      };
      
      return me;
   })();

   /**
    * Copy selected area into clipboard.
    * 
    * @param {number} begRow start row
    * @param {number} endRow end row
    * @param {number} begCol start column
    * @param {number} endCol end column
    * 
    * @returns {object} selected text.
   */
   me.copy = function(begRow, endRow, begCol, endCol)
   {
      return video.copy(begRow, endRow, begCol, endCol);
   };

   return me;
}());
