/*
** Module   : DojoToolkitHandler.java
** Abstract : delivers Dojo toolkit modules by loading them from the classpath as resources
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140109 First version.
** 002 MAG 20140701 Deliver multiple content types (MIME).
** 003 MAG 20140730 Change Dojo toolkit resource loading. Eliminate any logging messages.
** 004 GES 20150308 Moved to a more generic package location.
*/

package com.goldencode.p2j.web;

import java.io.*;
import java.net.*;

import javax.servlet.*;
import javax.servlet.http.*;

import org.eclipse.jetty.server.*;

/**
 * Jetty Handler which delivers Dojo toolkit modules.  The Dojo toolkit distribution zip file
 * should be located in the JVM CLASSPATH and modules are accessed as resources. The Dojo module
 * target request should start with DOJO_TOOLKIT_ROOT.
 * <p>
 * Starting from the target the Dojo module resource name is constructed by replacing the 
 * DOJO_TOOLKIT_ROOT with DOJO_TOOLKIT. Then the module resource is loaded and sent as response
 * to the requester.
 * <p>
 * IMPORTANT: Don't use logging anywhere inside this class. When used in a web client, this runs
 * in conjunction with an authentication plugin during session establishment with the P2J server
 * and using logging the application will deadlock.
*/
public class DojoToolkitHandler
extends AbstractHandlerCommons
{
   /** Dojo toolkit package. */
   private static final String DOJO_TOOLKIT = "dojo-release-1.10.0";

   /** Dojo toolkit root path. */
   private static final String DOJO_TOOLKIT_ROOT = "/dojo-toolkit";
      
   /**
    * Return a Dojo javascript file (module) if found or empty response and 
    * <code>HttpServletResponse.SC_NOT_FOUND</code> if the module was not found.
    *    
    * @param   target 
    *          The target of the request - either a URI or a name.
    * @param   base
    *          The base request.
    * @param   request
    *          The object or a wrapper of the request.
    * @param   response 
    *          The object or a wrapper of the response. 
    */
   public void handle(String              target, 
                      Request             base,
                      HttpServletRequest  request, 
                      HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      // check if this is a Dojo module request
      boolean isDojoModule = target.toLowerCase().startsWith(DOJO_TOOLKIT_ROOT);
      
      // for Dojo module requests, try to deliver the module 
      if (!base.isHandled() && isDojoModule)
      {
         // mark request as handled
         base.setHandled(true);
         deliverDojoModule(target, response);
      }    
   }
      
   /**
    * Deliver a Dojo toolkit module directly from zip file as a resource. The Dojo zip file MUST
    * be on CLASSPATH. In order to work properly this method MUST be synchronized.
    *
    * @param   module
    *          Dojo module path and name.
    * @param   response 
    *          The details and state of the response.
    *           
    * @throws  IOException
    *          ServletException
    *          If any error conditions occur.
    */
   private synchronized void deliverDojoModule(String module, HttpServletResponse response)
   throws IOException, 
          ServletException
   {
      String dojoModule = module.replace(DOJO_TOOLKIT_ROOT, DOJO_TOOLKIT);     

      // load resource
      URL url = Thread.currentThread().getContextClassLoader().getResource(dojoModule);

      if (url != null)
      {
         URLConnection conn = url.openConnection();
         sendResponse(response, conn, dojoModule);
      }
      else
      {
         // module not found
         response.setStatus(HttpServletResponse.SC_NOT_FOUND);                  
      }
   }
}
